@@ -3,7 +3,7 @@
  * Copyright (c) 2002 Juha Yrjölä
  * Copyright (c) 2002 Olaf Kirch
  * Copyright (c) 2003 Kevin Stefanik
- * Copyright (c) 2017 Michał Trojnara
+ * Copyright (c) 2016-2018 Michał Trojnara <Michal.Trojnara@stunnel.org>
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
@@ -30,6 +30,10 @@
 #include <stdio.h>
 #include <string.h>
 
+#if defined(_WIN32) || defined(_WIN64)
+#define strncasecmp _strnicmp
+#endif
+
 /* The maximum length of an internally-allocated PIN */
 #define MAX_PIN_LENGTH   32
 #define MAX_VALUE_LEN	200
@@ -49,6 +53,7 @@ struct st_engine_ctx {
 	char *init_args;
 	UI_METHOD *ui_method;
 	void *callback_data;
+	int force_login;
 
 	/* Engine initialization mutex */
 #if OPENSSL_VERSION_NUMBER >= 0x10100004L && !defined(LIBRESSL_VERSION_NUMBER)
@@ -67,11 +72,24 @@ struct st_engine_ctx {
 /* Utility functions                                                          */
 /******************************************************************************/
 
-static void dump_hex(FILE *stream, const char *val, const size_t len) {
+void ctx_log(ENGINE_CTX *ctx, int level, const char *format, ...)
+{
+	va_list ap;
+
+	if (level > ctx->verbose)
+			return;
+	va_start(ap, format);
+	vfprintf(stderr, format, ap);
+	va_end(ap);
+}
+
+static void dump_hex(ENGINE_CTX *ctx, int level,
+		const unsigned char *val, const size_t len)
+{
 	size_t n;
 
 	for (n = 0; n < len; n++)
-		fprintf(stream, "%02x", val[n]);
+		ctx_log(ctx, level, "%02x", val[n]);
 }
 
 /******************************************************************************/
@@ -101,7 +119,7 @@ static int ctx_get_pin(ENGINE_CTX *ctx, const char* token_label, UI_METHOD *ui_m
 	/* call ui to ask for a pin */
 	ui = UI_new_method(ui_method);
 	if (ui == NULL) {
-		fprintf(stderr, "UI_new failed\n");
+		ctx_log(ctx, 0, "UI_new failed\n");
 		return 0;
 	}
 	if (callback_data != NULL)
@@ -119,7 +137,7 @@ static int ctx_get_pin(ENGINE_CTX *ctx, const char* token_label, UI_METHOD *ui_m
 	}
 	if (!UI_dup_input_string(ui, prompt,
 			UI_INPUT_FLAG_DEFAULT_PWD, ctx->pin, 4, MAX_PIN_LENGTH)) {
-		fprintf(stderr, "UI_dup_input_string failed\n");
+		ctx_log(ctx, 0, "UI_dup_input_string failed\n");
 		UI_free(ui);
 		OPENSSL_free(prompt);
 		return 0;
@@ -127,7 +145,7 @@ static int ctx_get_pin(ENGINE_CTX *ctx, const char* token_label, UI_METHOD *ui_m
 	OPENSSL_free(prompt);
 
 	if (UI_process(ui)) {
-		fprintf(stderr, "UI_process failed\n");
+		ctx_log(ctx, 0, "UI_process failed\n");
 		UI_free(ui);
 		return 0;
 	}
@@ -135,6 +153,18 @@ static int ctx_get_pin(ENGINE_CTX *ctx, const char* token_label, UI_METHOD *ui_m
 	return 1;
 }
 
+/* Return 1 if the user has already logged in */
+static int slot_logged_in(ENGINE_CTX *ctx, PKCS11_SLOT *slot) {
+	int logged_in = 0;
+
+	/* Check if already logged in to avoid resetting state */
+	if (PKCS11_is_logged_in(slot, 0, &logged_in) != 0) {
+		ctx_log(ctx, 0, "Unable to check if already logged in\n");
+		return 0;
+	}
+	return logged_in;
+}
+
 /*
  * Log-into the token if necessary.
  *
@@ -147,17 +177,7 @@ static int ctx_get_pin(ENGINE_CTX *ctx, const char* token_label, UI_METHOD *ui_m
 static int ctx_login(ENGINE_CTX *ctx, PKCS11_SLOT *slot, PKCS11_TOKEN *tok,
 		UI_METHOD *ui_method, void *callback_data)
 {
-	int already_logged_in = 0;
-
-	if (!tok->loginRequired)
-		return 1;
-
-	/* Check if already logged in to avoid resetting state */
-	if (PKCS11_is_logged_in(slot, 0, &already_logged_in) != 0) {
-		fprintf(stderr, "Unable to check if already logged in\n");
-		return 0;
-	}
-	if (already_logged_in)
+	if (!(ctx->force_login || tok->loginRequired) || slot_logged_in(ctx, slot))
 		return 1;
 
 	/* If the token has a secure login (i.e., an external keypad),
@@ -170,13 +190,13 @@ static int ctx_login(ENGINE_CTX *ctx, PKCS11_SLOT *slot, PKCS11_TOKEN *tok,
 		ctx->pin = OPENSSL_malloc(MAX_PIN_LENGTH+1);
 		ctx->pin_length = MAX_PIN_LENGTH;
 		if (ctx->pin == NULL) {
-			fprintf(stderr, "Could not allocate memory for PIN\n");
+			ctx_log(ctx, 0, "Could not allocate memory for PIN\n");
 			return 0;
 		}
 		memset(ctx->pin, 0, MAX_PIN_LENGTH+1);
 		if (!ctx_get_pin(ctx, tok->label, ui_method, callback_data)) {
 			ctx_destroy_pin(ctx);
-			fprintf(stderr, "No PIN code was entered\n");
+			ctx_log(ctx, 0, "No PIN code was entered\n");
 			return 0;
 		}
 	}
@@ -185,7 +205,7 @@ static int ctx_login(ENGINE_CTX *ctx, PKCS11_SLOT *slot, PKCS11_TOKEN *tok,
 	if (PKCS11_login(slot, 0, ctx->pin)) {
 		/* Login failed, so free the PIN if present */
 		ctx_destroy_pin(ctx);
-		fprintf(stderr, "Login failed\n");
+		ctx_log(ctx, 0, "Login failed\n");
 		return 0;
 	}
 	return 1;
@@ -230,7 +250,6 @@ ENGINE_CTX *ctx_new()
 int ctx_destroy(ENGINE_CTX *ctx)
 {
 	if (ctx) {
-		ctx_finish(ctx);
 		ctx_destroy_pin(ctx);
 		OPENSSL_free(ctx->module);
 		OPENSSL_free(ctx->init_args);
@@ -252,8 +271,7 @@ static void ctx_init_libp11_unlocked(ENGINE_CTX *ctx)
 	PKCS11_SLOT *slot_list = NULL;
 	unsigned int slot_count = 0;
 
-	if (ctx->verbose)
-		fprintf(stderr, "PKCS#11: Initializing the engine\n");
+	ctx_log(ctx, 1, "PKCS#11: Initializing the engine\n");
 
 	pkcs11_ctx = PKCS11_CTX_new();
 	PKCS11_CTX_init_args(pkcs11_ctx, ctx->init_args);
@@ -261,22 +279,21 @@ static void ctx_init_libp11_unlocked(ENGINE_CTX *ctx)
 
 	/* PKCS11_CTX_load() uses C_GetSlotList() via p11-kit */
 	if (PKCS11_CTX_load(pkcs11_ctx, ctx->module) < 0) {
-		fprintf(stderr, "Unable to load module %s\n", ctx->module);
+		ctx_log(ctx, 0, "Unable to load module %s\n", ctx->module);
 		PKCS11_CTX_free(pkcs11_ctx);
 		return;
 	}
 
 	/* PKCS11_enumerate_slots() uses C_GetSlotList() via libp11 */
 	if (PKCS11_enumerate_slots(pkcs11_ctx, &slot_list, &slot_count) < 0) {
-		fprintf(stderr, "Failed to enumerate slots\n");
+		ctx_log(ctx, 0, "Failed to enumerate slots\n");
 		PKCS11_CTX_unload(pkcs11_ctx);
 		PKCS11_CTX_free(pkcs11_ctx);
 		return;
 	}
 
-	if (ctx->verbose)
-		fprintf(stderr, "Found %u slot%s\n", slot_count,
-			slot_count <= 1 ? "" : "s");
+	ctx_log(ctx, 1, "Found %u slot%s\n", slot_count,
+		slot_count <= 1 ? "" : "s");
 
 	ctx->pkcs11_ctx = pkcs11_ctx;
 	ctx->slot_list = slot_list;
@@ -322,6 +339,8 @@ int ctx_init(ENGINE_CTX *ctx)
 		CRYPTO_w_lock(CRYPTO_LOCK_ENGINE);
 		return ctx->pkcs11_ctx && ctx->slot_list ? 1 : 0;
 	}
+#else
+	(void)ctx; /* squash the unused parameter warning */
 #endif
 	return 1;
 }
@@ -337,11 +356,7 @@ int ctx_finish(ENGINE_CTX *ctx)
 			ctx->slot_count = 0;
 		}
 		if (ctx->pkcs11_ctx) {
-			/* Modules cannot be unloaded in pkcs11_finish() nor
-			 * ctx_destroy() because of a deadlock in PKCS#11
-			 * modules that internally use OpenSSL engines.
-			 * A memory leak is better than a deadlock... */
-			/* PKCS11_CTX_unload(ctx->pkcs11_ctx); */
+			PKCS11_CTX_unload(ctx->pkcs11_ctx);
 			PKCS11_CTX_free(ctx->pkcs11_ctx);
 			ctx->pkcs11_ctx = NULL;
 		}
@@ -377,11 +392,20 @@ static X509 *ctx_load_cert(ENGINE_CTX *ctx, const char *s_slot_cert_id,
 		return NULL;
 
 	if (s_slot_cert_id && *s_slot_cert_id) {
-		if (!strncmp(s_slot_cert_id, "pkcs11:", 7)) {
-			n = parse_pkcs11_uri(s_slot_cert_id, &match_tok,
+		if (!strncasecmp(s_slot_cert_id, "pkcs11:", 7)) {
+			n = parse_pkcs11_uri(ctx, s_slot_cert_id, &match_tok,
 				cert_id, &cert_id_len,
 				tmp_pin, &tmp_pin_len, &cert_label);
-			if (n && tmp_pin_len > 0 && tmp_pin[0] != 0) {
+			if (!n) {
+				ctx_log(ctx, 0,
+					"The certificate ID is not a valid PKCS#11 URI\n"
+					"The PKCS#11 URI format is defined by RFC7512\n");
+				ENGerr(ENG_F_CTX_LOAD_CERT, ENG_R_INVALID_ID);
+				return NULL;
+			}
+			if (tmp_pin_len > 0 && tmp_pin[0] != 0) {
+				if (!login)
+					return NULL; /* Process on second attempt */
 				ctx_destroy_pin(ctx);
 				ctx->pin = OPENSSL_malloc(MAX_PIN_LENGTH+1);
 				if (ctx->pin != NULL) {
@@ -390,39 +414,30 @@ static X509 *ctx_load_cert(ENGINE_CTX *ctx, const char *s_slot_cert_id,
 					ctx->pin_length = tmp_pin_len;
 				}
 			}
-
-			if (!n) {
-				fprintf(stderr,
-					"The certificate ID is not a valid PKCS#11 URI\n"
-					"The PKCS#11 URI format is defined by RFC7512\n");
-				return NULL;
-			}
 		} else {
-			n = parse_slot_id_string(s_slot_cert_id, &slot_nr,
+			n = parse_slot_id_string(ctx, s_slot_cert_id, &slot_nr,
 				cert_id, &cert_id_len, &cert_label);
-
 			if (!n) {
-				fprintf(stderr,
+				ctx_log(ctx, 0,
 					"The certificate ID is not a valid PKCS#11 URI\n"
 					"The PKCS#11 URI format is defined by RFC7512\n"
 					"The legacy ENGINE_pkcs11 ID format is also "
 					"still accepted for now\n");
+				ENGerr(ENG_F_CTX_LOAD_CERT, ENG_R_INVALID_ID);
 				return NULL;
 			}
 		}
-		if (ctx->verbose) {
-			fprintf(stderr, "Looking in slot %d for certificate: ",
-				slot_nr);
-			if (cert_id_len != 0) {
-				fprintf(stderr, "id=");
-				dump_hex(stderr, cert_id, cert_id_len);
-			}
-			if (cert_id_len != 0 && cert_label != NULL)
-				fprintf(stderr, " ");
-			if (cert_label != NULL)
-				fprintf(stderr, "label=%s", cert_label);
-			fprintf(stderr, "\n");
+		ctx_log(ctx, 1, "Looking in slot %d for certificate: ",
+			slot_nr);
+		if (cert_id_len != 0) {
+			ctx_log(ctx, 1, "id=");
+			dump_hex(ctx, 1, cert_id, cert_id_len);
 		}
+		if (cert_id_len != 0 && cert_label != NULL)
+			ctx_log(ctx, 1, " ");
+		if (cert_label != NULL)
+			ctx_log(ctx, 1, "label=%s", cert_label);
+		ctx_log(ctx, 1, "\n");
 	}
 
 	for (n = 0; n < ctx->slot_count; n++) {
@@ -459,17 +474,15 @@ static X509 *ctx_load_cert(ENGINE_CTX *ctx, const char *s_slot_cert_id,
 					!strcmp(match_tok->model, slot->token->model))) {
 			found_slot = slot;
 		}
-		if (ctx->verbose) {
-			fprintf(stderr, "[%lu] %-25.25s  %-16s",
-				PKCS11_get_slotid_from_slot(slot),
-				slot->description, flags);
-			if (slot->token) {
-				fprintf(stderr, "  (%s)",
-					slot->token->label[0] ?
-					slot->token->label : "no label");
-			}
-			fprintf(stderr, "\n");
+		ctx_log(ctx, 1, "[%lu] %-25.25s  %-16s",
+			PKCS11_get_slotid_from_slot(slot),
+			slot->description, flags);
+		if (slot->token) {
+			ctx_log(ctx, 1, "  (%s)",
+				slot->token->label[0] ?
+				slot->token->label : "no label");
 		}
+		ctx_log(ctx, 1, "\n");
 	}
 
 	if (match_tok) {
@@ -482,46 +495,42 @@ static X509 *ctx_load_cert(ENGINE_CTX *ctx, const char *s_slot_cert_id,
 	if (found_slot) {
 		slot = found_slot;
 	} else if (match_tok) {
-		fprintf(stderr, "Specified object not found\n");
+		ctx_log(ctx, 0, "Specified object not found\n");
 		return NULL;
 	} else if (slot_nr == -1) {
 		if (!(slot = PKCS11_find_token(ctx->pkcs11_ctx,
 				ctx->slot_list, ctx->slot_count))) {
-			fprintf(stderr, "No tokens found\n");
+			ctx_log(ctx, 0, "No tokens found\n");
 			return NULL;
 		}
 	} else {
-		fprintf(stderr, "Invalid slot number: %d\n", slot_nr);
+		ctx_log(ctx, 0, "Invalid slot number: %d\n", slot_nr);
 		return NULL;
 	}
 	tok = slot->token;
 
 	if (tok == NULL) {
-		fprintf(stderr, "Empty token found\n");
+		ctx_log(ctx, 0, "Empty token found\n");
 		return NULL;
 	}
 
-	if (ctx->verbose) {
-		fprintf(stderr, "Found slot:  %s\n", slot->description);
-		fprintf(stderr, "Found token: %s\n", slot->token->label);
-	}
+	ctx_log(ctx, 1, "Found slot:  %s\n", slot->description);
+	ctx_log(ctx, 1, "Found token: %s\n", slot->token->label);
 
 	/* In several tokens certificates are marked as private */
 	if (login && !ctx_login(ctx, slot, tok,
 			ctx->ui_method, ctx->callback_data)) {
-		fprintf(stderr, "Login to token failed, returning NULL...\n");
+		ctx_log(ctx, 0, "Login to token failed, returning NULL...\n");
 		return NULL;
 	}
 
 	if (PKCS11_enumerate_certs(tok, &certs, &cert_count)) {
-		fprintf(stderr, "Unable to enumerate certificates\n");
+		ctx_log(ctx, 0, "Unable to enumerate certificates\n");
 		return NULL;
 	}
 
-	if (ctx->verbose) {
-		fprintf(stderr, "Found %u cert%s:\n", cert_count,
-			(cert_count <= 1) ? "" : "s");
-	}
+	ctx_log(ctx, 1, "Found %u cert%s:\n", cert_count,
+		(cert_count <= 1) ? "" : "s");
 	if ((s_slot_cert_id && *s_slot_cert_id) &&
 			(cert_id_len != 0 || cert_label != NULL)) {
 		for (n = 0; n < cert_count; n++) {
@@ -533,15 +542,23 @@ static X509 *ctx_load_cert(ENGINE_CTX *ctx, const char *s_slot_cert_id,
 					memcmp(k->id, cert_id, cert_id_len) == 0)
 				selected_cert = k;
 		}
-	} else {
-		selected_cert = certs; /* Use the first certificate */
+	} else { 
+		for (n = 0; n < cert_count; n++) {
+			PKCS11_CERT *k = certs + n;
+			if(k->id && *(k->id)) {
+				selected_cert = k; /* Use the first certificate with nonempty id */
+				break;
+			}
+		}
+		if(!selected_cert)
+			selected_cert = certs; /* Use the first certificate */
 	}
 
 	if (selected_cert != NULL) {
 		x509 = X509_dup(selected_cert->x509);
 	} else {
 		if (login) /* Only print the error on the second attempt */
-			fprintf(stderr, "Certificate not found.\n");
+			ctx_log(ctx, 0, "Certificate not found.\n");
 		x509 = NULL;
 	}
 	if (cert_label != NULL)
@@ -556,15 +573,26 @@ static int ctx_ctrl_load_cert(ENGINE_CTX *ctx, void *p)
 		X509 *cert;
 	} *parms = p;
 
-	if (parms->cert != NULL)
+	if (parms == NULL) {
+		ENGerr(ENG_F_CTX_CTRL_LOAD_CERT, ERR_R_PASSED_NULL_PARAMETER);
 		return 0;
-
-	parms->cert = ctx_load_cert(ctx, parms->s_slot_cert_id, 0);
-	if (parms->cert == NULL) /* Try again with login */
+	}
+	if (parms->cert != NULL) {
+		ENGerr(ENG_F_CTX_CTRL_LOAD_CERT, ENG_R_INVALID_PARAMETER);
+		return 0;
+	}
+	ERR_clear_error();
+	if (!ctx->force_login)
+		parms->cert = ctx_load_cert(ctx, parms->s_slot_cert_id, 0);
+	if (parms->cert == NULL) { /* Try again with login */
+		ERR_clear_error();
 		parms->cert = ctx_load_cert(ctx, parms->s_slot_cert_id, 1);
-
-	if (parms->cert == NULL)
+	}
+	if (parms->cert == NULL) {
+		if (!ERR_peek_last_error())
+			ENGerr(ENG_F_CTX_CTRL_LOAD_CERT, ENG_R_OBJECT_NOT_FOUND);
 		return 0;
+	}
 	return 1;
 }
 
@@ -580,9 +608,8 @@ static EVP_PKEY *ctx_load_key(ENGINE_CTX *ctx, const char *s_slot_key_id,
 	PKCS11_SLOT *found_slot = NULL;
 	PKCS11_TOKEN *tok, *match_tok = NULL;
 	PKCS11_KEY *keys, *selected_key = NULL;
-	PKCS11_CERT *certs;
-	EVP_PKEY *pk;
-	unsigned int cert_count, key_count, n, m;
+	EVP_PKEY *pk = NULL;
+	unsigned int key_count, n, m;
 	unsigned char key_id[MAX_VALUE_LEN / 2];
 	size_t key_id_len = sizeof(key_id);
 	char *key_label = NULL;
@@ -592,19 +619,26 @@ static EVP_PKEY *ctx_load_key(ENGINE_CTX *ctx, const char *s_slot_key_id,
 	char flags[64];
 
 	if (ctx_init_libp11(ctx)) /* Delayed libp11 initialization */
-		return NULL;
+		goto error;
 
-	if (ctx->verbose)
-		fprintf(stderr, "Loading %s key \"%s\"\n",
-			(char *)(isPrivate ? "private" : "public"),
-			s_slot_key_id);
+	ctx_log(ctx, 1, "Loading %s key \"%s\"\n",
+		(char *)(isPrivate ? "private" : "public"),
+		s_slot_key_id);
 	if (s_slot_key_id && *s_slot_key_id) {
-		if (!strncmp(s_slot_key_id, "pkcs11:", 7)) {
-			n = parse_pkcs11_uri(s_slot_key_id, &match_tok,
+		if (!strncasecmp(s_slot_key_id, "pkcs11:", 7)) {
+			n = parse_pkcs11_uri(ctx, s_slot_key_id, &match_tok,
 				key_id, &key_id_len,
 				tmp_pin, &tmp_pin_len, &key_label);
-
-			if (n && tmp_pin_len > 0 && tmp_pin[0] != 0) {
+			if (!n) {
+				ctx_log(ctx, 0,
+					"The key ID is not a valid PKCS#11 URI\n"
+					"The PKCS#11 URI format is defined by RFC7512\n");
+				ENGerr(ENG_F_CTX_LOAD_KEY, ENG_R_INVALID_ID);
+				goto error;
+			}
+			if (tmp_pin_len > 0 && tmp_pin[0] != 0) {
+				if (!login)
+					goto error; /* Process on second attempt */
 				ctx_destroy_pin(ctx);
 				ctx->pin = OPENSSL_malloc(MAX_PIN_LENGTH+1);
 				if (ctx->pin != NULL) {
@@ -613,39 +647,30 @@ static EVP_PKEY *ctx_load_key(ENGINE_CTX *ctx, const char *s_slot_key_id,
 					ctx->pin_length = tmp_pin_len;
 				}
 			}
-
-			if (!n) {
-				fprintf(stderr,
-					"The certificate ID is not a valid PKCS#11 URI\n"
-					"The PKCS#11 URI format is defined by RFC7512\n");
-				return NULL;
-			}
 		} else {
-			n = parse_slot_id_string(s_slot_key_id, &slot_nr,
+			n = parse_slot_id_string(ctx, s_slot_key_id, &slot_nr,
 				key_id, &key_id_len, &key_label);
-
 			if (!n) {
-				fprintf(stderr,
-					"The certificate ID is not a valid PKCS#11 URI\n"
+				ctx_log(ctx, 0,
+					"The key ID is not a valid PKCS#11 URI\n"
 					"The PKCS#11 URI format is defined by RFC7512\n"
 					"The legacy ENGINE_pkcs11 ID format is also "
 					"still accepted for now\n");
-				return NULL;
+				ENGerr(ENG_F_CTX_LOAD_KEY, ENG_R_INVALID_ID);
+				goto error;
 			}
 		}
-		if (ctx->verbose) {
-			fprintf(stderr, "Looking in slot %d for key: ",
-				slot_nr);
-			if (key_id_len != 0) {
-				fprintf(stderr, "id=");
-				dump_hex(stderr, key_id, key_id_len);
-			}
-			if (key_id_len != 0 && key_label != NULL)
-				fprintf(stderr, " ");
-			if (key_label != NULL)
-				fprintf(stderr, "label=%s", key_label);
-			fprintf(stderr, "\n");
+		ctx_log(ctx, 1, "Looking in slot %d for key: ",
+			slot_nr);
+		if (key_id_len != 0) {
+			ctx_log(ctx, 1, "id=");
+			dump_hex(ctx, 1, key_id, key_id_len);
 		}
+		if (key_id_len != 0 && key_label != NULL)
+			ctx_log(ctx, 1, " ");
+		if (key_label != NULL)
+			ctx_log(ctx, 1, "label=%s", key_label);
+		ctx_log(ctx, 1, "\n");
 	}
 
 	for (n = 0; n < ctx->slot_count; n++) {
@@ -682,17 +707,15 @@ static EVP_PKEY *ctx_load_key(ENGINE_CTX *ctx, const char *s_slot_key_id,
 					!strcmp(match_tok->model, slot->token->model))) {
 			found_slot = slot;
 		}
-		if (ctx->verbose) {
-			fprintf(stderr, "[%lu] %-25.25s  %-16s",
-				PKCS11_get_slotid_from_slot(slot),
-				slot->description, flags);
-			if (slot->token) {
-				fprintf(stderr, "  (%s)",
-					slot->token->label[0] ?
-					slot->token->label : "no label");
-			}
-			fprintf(stderr, "\n");
+		ctx_log(ctx, 1, "[%lu] %-25.25s  %-16s",
+			PKCS11_get_slotid_from_slot(slot),
+			slot->description, flags);
+		if (slot->token) {
+			ctx_log(ctx, 1, "  (%s)",
+				slot->token->label[0] ?
+				slot->token->label : "no label");
 		}
+		ctx_log(ctx, 1, "\n");
 	}
 
 	if (match_tok) {
@@ -705,105 +728,76 @@ static EVP_PKEY *ctx_load_key(ENGINE_CTX *ctx, const char *s_slot_key_id,
 	if (found_slot) {
 		slot = found_slot;
 	} else if (match_tok) {
-		fprintf(stderr, "Specified object not found\n");
-		return NULL;
+		ctx_log(ctx, 0, "Specified object not found\n");
+		goto error;
 	} else if (slot_nr == -1) {
 		if (!(slot = PKCS11_find_token(ctx->pkcs11_ctx,
 				ctx->slot_list, ctx->slot_count))) {
-			fprintf(stderr, "No tokens found\n");
-			return NULL;
+			ctx_log(ctx, 0, "No tokens found\n");
+			goto error;
 		}
 	} else {
-		fprintf(stderr, "Invalid slot number: %d\n", slot_nr);
-		return NULL;
+		ctx_log(ctx, 0, "Invalid slot number: %d\n", slot_nr);
+		goto error;
 	}
 	tok = slot->token;
 
 	if (tok == NULL) {
-		fprintf(stderr, "Found empty token\n");
-		return NULL;
+		ctx_log(ctx, 0, "Found empty token\n");
+		goto error;
 	}
 	/* The following check is non-critical to ensure interoperability
 	 * with some other (which ones?) PKCS#11 libraries */
 	if (!tok->initialized)
-		fprintf(stderr, "Found uninitialized token\n");
+		ctx_log(ctx, 0, "Found uninitialized token\n");
 	if (isPrivate && !tok->userPinSet && !tok->readOnly) {
-		fprintf(stderr, "Found slot without user PIN\n");
-		return NULL;
+		ctx_log(ctx, 0, "Found slot without user PIN\n");
+		goto error;
 	}
 
-	if (ctx->verbose) {
-		fprintf(stderr, "Found slot:  %s\n", slot->description);
-		fprintf(stderr, "Found token: %s\n", slot->token->label);
-	}
-
-	if (PKCS11_enumerate_certs(tok, &certs, &cert_count)) {
-		fprintf(stderr, "Unable to enumerate certificates\n");
-		return NULL;
-	}
-
-	if (ctx->verbose) {
-		fprintf(stderr, "Found %u certificate%s:\n", cert_count,
-			(cert_count <= 1) ? "" : "s");
-		for (n = 0; n < cert_count; n++) {
-			PKCS11_CERT *c = certs + n;
-			char *dn = NULL;
-
-			fprintf(stderr, "  %2u    id=", n + 1);
-			dump_hex(stderr, c->id, c->id_len);
-			fprintf(stderr, " label=%s", c->label);
-			if (c->x509)
-				dn = X509_NAME_oneline(X509_get_subject_name(c->x509), NULL, 0);
-			if (dn) {
-				fprintf(stderr, " (%s)", dn);
-				OPENSSL_free(dn);
-			}
-			fprintf(stderr, "\n");
-		}
-	}
+	ctx_log(ctx, 1, "Found slot:  %s\n", slot->description);
+	ctx_log(ctx, 1, "Found token: %s\n", slot->token->label);
 
 	/* Both private and public keys can have the CKA_PRIVATE attribute
 	 * set and thus require login (even to retrieve attributes!) */
 	if (login && !ctx_login(ctx, slot, tok, ui_method, callback_data)) {
-		fprintf(stderr, "Login to token failed, returning NULL...\n");
-		return NULL;
+		ctx_log(ctx, 0, "Login to token failed, returning NULL...\n");
+		goto error;
 	}
 
 	if (isPrivate) {
 		/* Make sure there is at least one private key on the token */
 		if (PKCS11_enumerate_keys(tok, &keys, &key_count)) {
-			fprintf(stderr, "Unable to enumerate private keys\n");
-			return NULL;
+			ctx_log(ctx, 0, "Unable to enumerate private keys\n");
+			goto error;
 		}
 	} else {
 		/* Make sure there is at least one public key on the token */
 		if (PKCS11_enumerate_public_keys(tok, &keys, &key_count)) {
-			fprintf(stderr, "Unable to enumerate public keys\n");
-			return NULL;
+			ctx_log(ctx, 0, "Unable to enumerate public keys\n");
+			goto error;
 		}
 	}
 	if (key_count == 0) {
-		fprintf(stderr, "No %s keys found.\n",
-			(char *)(isPrivate ? "private" : "public"));
-		return NULL;
+		if (login) /* Only print the error on the second attempt */
+			ctx_log(ctx, 0, "No %s keys found.\n",
+				(char *)(isPrivate ? "private" : "public"));
+		goto error;
 	}
-	if (ctx->verbose)
-		fprintf(stderr, "Found %u %s key%s:\n", key_count,
-			(char *)(isPrivate ? "private" : "public"),
-			(key_count == 1) ? "" : "s");
+	ctx_log(ctx, 1, "Found %u %s key%s:\n", key_count,
+		(char *)(isPrivate ? "private" : "public"),
+		(key_count == 1) ? "" : "s");
 
 	if (s_slot_key_id && *s_slot_key_id &&
 			(key_id_len != 0 || key_label != NULL)) {
 		for (n = 0; n < key_count; n++) {
 			PKCS11_KEY *k = keys + n;
 
-			if (ctx->verbose) {
-				fprintf(stderr, "  %2u %c%c id=", n + 1,
-					k->isPrivate ? 'P' : ' ',
-					k->needLogin ? 'L' : ' ');
-				dump_hex(stderr, k->id, k->id_len);
-				fprintf(stderr, " label=%s\n", k->label);
-			}
+			ctx_log(ctx, 1, "  %2u %c%c id=", n + 1,
+				k->isPrivate ? 'P' : ' ',
+				k->needLogin ? 'L' : ' ');
+			dump_hex(ctx, 1, k->id, k->id_len);
+			ctx_log(ctx, 1, " label=%s\n", k->label);
 			if (key_label != NULL && strcmp(k->label, key_label) == 0)
 				selected_key = k;
 			if (key_id_len != 0 && k->id_len == key_id_len
@@ -820,9 +814,10 @@ static EVP_PKEY *ctx_load_key(ENGINE_CTX *ctx, const char *s_slot_key_id,
 			PKCS11_get_public_key(selected_key);
 	} else {
 		if (login) /* Only print the error on the second attempt */
-			fprintf(stderr, "Key not found.\n");
+			ctx_log(ctx, 0, "Key not found.\n");
 		pk = NULL;
 	}
+error:
 	if (key_label != NULL)
 		OPENSSL_free(key_label);
 	return pk;
@@ -831,13 +826,19 @@ static EVP_PKEY *ctx_load_key(ENGINE_CTX *ctx, const char *s_slot_key_id,
 EVP_PKEY *ctx_load_pubkey(ENGINE_CTX *ctx, const char *s_key_id,
 		UI_METHOD *ui_method, void *callback_data)
 {
-	EVP_PKEY *pk;
+	EVP_PKEY *pk = NULL;
 
-	pk = ctx_load_key(ctx, s_key_id, ui_method, callback_data, 0, 0);
-	if (pk == NULL) /* Try again with login */
+	ERR_clear_error();
+	if (!ctx->force_login)
+		pk = ctx_load_key(ctx, s_key_id, ui_method, callback_data, 0, 0);
+	if (pk == NULL) { /* Try again with login */
+		ERR_clear_error();
 		pk = ctx_load_key(ctx, s_key_id, ui_method, callback_data, 0, 1);
+	}
 	if (pk == NULL) {
-		fprintf(stderr, "PKCS11_load_public_key returned NULL\n");
+		ctx_log(ctx, 0, "PKCS11_load_public_key returned NULL\n");
+		if (!ERR_peek_last_error())
+			ENGerr(ENG_F_CTX_LOAD_PUBKEY, ENG_R_OBJECT_NOT_FOUND);
 		return NULL;
 	}
 	return pk;
@@ -846,13 +847,19 @@ EVP_PKEY *ctx_load_pubkey(ENGINE_CTX *ctx, const char *s_key_id,
 EVP_PKEY *ctx_load_privkey(ENGINE_CTX *ctx, const char *s_key_id,
 		UI_METHOD *ui_method, void *callback_data)
 {
-	EVP_PKEY *pk;
+	EVP_PKEY *pk = NULL;
 
-	pk = ctx_load_key(ctx, s_key_id, ui_method, callback_data, 1, 0);
-	if (pk == NULL) /* Try again with login */
+	ERR_clear_error();
+	if (!ctx->force_login)
+		pk = ctx_load_key(ctx, s_key_id, ui_method, callback_data, 1, 0);
+	if (pk == NULL) { /* Try again with login */
+		ERR_clear_error();
 		pk = ctx_load_key(ctx, s_key_id, ui_method, callback_data, 1, 1);
+	}
 	if (pk == NULL) {
-		fprintf(stderr, "PKCS11_get_private_key returned NULL\n");
+		ctx_log(ctx, 0, "PKCS11_get_private_key returned NULL\n");
+		if (!ERR_peek_last_error())
+			ENGerr(ENG_F_CTX_LOAD_PRIVKEY, ENG_R_OBJECT_NOT_FOUND);
 		return NULL;
 	}
 	return pk;
@@ -886,6 +893,7 @@ static int ctx_ctrl_set_pin(ENGINE_CTX *ctx, const char *pin)
 {
 	/* Pre-condition check */
 	if (pin == NULL) {
+		ENGerr(ENG_F_CTX_CTRL_SET_PIN, ERR_R_PASSED_NULL_PARAMETER);
 		errno = EINVAL;
 		return 0;
 	}
@@ -894,10 +902,13 @@ static int ctx_ctrl_set_pin(ENGINE_CTX *ctx, const char *pin)
 	 * shall be returned and errno shall be set. */
 	ctx_destroy_pin(ctx);
 	ctx->pin = OPENSSL_strdup(pin);
-	if (ctx->pin != NULL)
-		ctx->pin_length = strlen(ctx->pin);
-
-	return ctx->pin != NULL;
+	if (ctx->pin == NULL) {
+		ENGerr(ENG_F_CTX_CTRL_SET_PIN, ERR_R_MALLOC_FAILURE);
+		errno = ENOMEM;
+		return 0;
+	}
+	ctx->pin_length = strlen(ctx->pin);
+	return 1;
 }
 
 static int ctx_ctrl_inc_verbose(ENGINE_CTX *ctx)
@@ -906,6 +917,12 @@ static int ctx_ctrl_inc_verbose(ENGINE_CTX *ctx)
 	return 1;
 }
 
+static int ctx_ctrl_set_quiet(ENGINE_CTX *ctx)
+{
+	ctx->verbose = -1;
+	return 1;
+}
+
 static int ctx_ctrl_set_init_args(ENGINE_CTX *ctx, const char *init_args_orig)
 {
 	OPENSSL_free(ctx->init_args);
@@ -931,6 +948,12 @@ static int ctx_ctrl_set_callback_data(ENGINE_CTX *ctx, void *callback_data)
 	return 1;
 }
 
+static int ctx_ctrl_force_login(ENGINE_CTX *ctx)
+{
+	ctx->force_login = 1;
+	return 1;
+}
+
 int ctx_engine_ctrl(ENGINE_CTX *ctx, int cmd, long i, void *p, void (*f)())
 {
 	(void)i; /* We don't currently take integer parameters */
@@ -943,6 +966,8 @@ int ctx_engine_ctrl(ENGINE_CTX *ctx, int cmd, long i, void *p, void (*f)())
 		return ctx_ctrl_set_pin(ctx, (const char *)p);
 	case CMD_VERBOSE:
 		return ctx_ctrl_inc_verbose(ctx);
+	case CMD_QUIET:
+		return ctx_ctrl_set_quiet(ctx);
 	case CMD_LOAD_CERT_CTRL:
 		return ctx_ctrl_load_cert(ctx, p);
 	case CMD_INIT_ARGS:
@@ -953,7 +978,10 @@ int ctx_engine_ctrl(ENGINE_CTX *ctx, int cmd, long i, void *p, void (*f)())
 	case ENGINE_CTRL_SET_CALLBACK_DATA:
 	case CMD_SET_CALLBACK_DATA:
 		return ctx_ctrl_set_callback_data(ctx, p);
+	case CMD_FORCE_LOGIN:
+		return ctx_ctrl_force_login(ctx);
 	default:
+		ENGerr(ENG_F_CTX_ENGINE_CTRL, ENG_R_UNKNOWN_COMMAND);
 		break;
 	}
 	return 0;
