@@ -2,7 +2,7 @@
  * Copyright (C) 2005 Olaf Kirch <okir@lst.de>
  * Copyright (C) 2011, 2013 Douglas E. Engert <deengert@anl.gov>
  * Copyright (C) 2014, 2016 Douglas E. Engert <deengert@gmail.com>
- * Copyright (C) 2016 Michał Trojnara <Michal.Trojnara@stunnel.org>
+ * Copyright (C) 2016-2018 Michał Trojnara <Michal.Trojnara@stunnel.org>
  *
  *  This library is free software; you can redistribute it and/or
  *  modify it under the terms of the GNU Lesser General Public
@@ -37,6 +37,8 @@
 #include <openssl/ecdh.h>
 #endif
 
+#ifndef OPENSSL_NO_EC
+
 #if OPENSSL_VERSION_NUMBER >= 0x10100004L && !defined(LIBRESSL_VERSION_NUMBER)
 typedef int (*compute_key_fn)(unsigned char **, size_t *,
 	const EC_POINT *, const EC_KEY *);
@@ -49,8 +51,6 @@ static compute_key_fn ossl_ecdh_compute_key;
 
 static int ec_ex_index = 0;
 
-#ifndef OPENSSL_NO_EC
-
 /********** Missing ECDSA_METHOD functions for OpenSSL < 1.1.0 */
 
 typedef ECDSA_SIG *(*sign_sig_fn)(const unsigned char *, int,
@@ -184,12 +184,58 @@ static void free_ec_ex_index()
 
 /********** EVP_PKEY retrieval */
 
+/* Retrieve EC parameters from key into ec
+ * return nonzero on error */
+static int pkcs11_get_params(EC_KEY *ec, PKCS11_KEY *key)
+{
+	CK_BYTE *params;
+	size_t params_len = 0;
+	const unsigned char *a;
+	int rv;
+
+	if (key_getattr_alloc(key, CKA_EC_PARAMS, &params, &params_len))
+		return -1;
+
+	a = params;
+	rv = d2i_ECParameters(&ec, &a, (long)params_len) == NULL;
+	OPENSSL_free(params);
+	return rv;
+}
+
+/* Retrieve EC point from key into ec
+ * return nonzero on error */
+static int pkcs11_get_point(EC_KEY *ec, PKCS11_KEY *key)
+{
+	CK_BYTE *point;
+	size_t point_len = 0;
+	const unsigned char *a;
+	ASN1_OCTET_STRING *os;
+	int rv = -1;
+
+	if (key == NULL ||
+			key_getattr_alloc(key, CKA_EC_POINT, &point, &point_len))
+		return -1;
+
+	/* PKCS#11-compliant modules should return ASN1_OCTET_STRING */
+	a = point;
+	os = d2i_ASN1_OCTET_STRING(NULL, &a, (long)point_len);
+	if (os) {
+		a = os->data;
+		rv = o2i_ECPublicKey(&ec, &a, os->length) == NULL;
+		ASN1_STRING_free(os);
+	}
+	if (rv) { /* Workaround for broken PKCS#11 modules */
+		a = point;
+		rv = o2i_ECPublicKey(&ec, &a, (long)point_len) == NULL;
+	}
+	OPENSSL_free(point);
+	return rv;
+}
+
 static EC_KEY *pkcs11_get_ec(PKCS11_KEY *key)
 {
-	EC_KEY *ec, *found_params = NULL, *found_point = NULL;
-	CK_BYTE *params, *point;
-	size_t params_len = 0, point_len = 0;
-	PKCS11_KEY *pubkey;
+	EC_KEY *ec;
+	int no_params, no_point;
 
 	ec = EC_KEY_new();
 	if (ec == NULL)
@@ -200,39 +246,13 @@ static EC_KEY *pkcs11_get_ec(PKCS11_KEY *key)
 	 * Continue even if it fails, as the sign operation does not need
 	 * it if the PKCS#11 module or the hardware can figure this out
 	 */
-	if (!key_getattr_alloc(key, CKA_EC_PARAMS, &params, &params_len)) {
-		const unsigned char *a = params;
-
-		/* Convert to OpenSSL parmas */
-		found_params = d2i_ECParameters(&ec, &a, (long)params_len);
-		OPENSSL_free(params);
-	}
-
-	/* Now retrieve the point */
-	pubkey = key->isPrivate ? pkcs11_find_key_from_key(key) : key;
-	if (pubkey == NULL)
-		return ec;
-	if (!key_getattr_alloc(pubkey, CKA_EC_POINT, &point, &point_len)) {
-		const unsigned char *a;
-		ASN1_OCTET_STRING *os;
-
-		/* PKCS#11-compliant modules should return ASN1_OCTET_STRING */
-		a = point;
-		os = d2i_ASN1_OCTET_STRING(NULL, &a, (long)point_len);
-		if (os) {
-			a = os->data;
-			found_point = o2i_ECPublicKey(&ec, &a, os->length);
-			ASN1_STRING_free(os);
-		}
-		if (found_point == NULL) { /* Workaround for broken PKCS#11 modules */
-			a = point;
-			found_point = o2i_ECPublicKey(&ec, &a, point_len);
-		}
-		OPENSSL_free(point);
-	}
+	no_params = pkcs11_get_params(ec, key);
+	no_point = pkcs11_get_point(ec, key);
+	if (no_point && key->isPrivate) /* Retry with the public key */
+		no_point = pkcs11_get_point(ec, pkcs11_find_key_from_key(key));
 
 	/* A public keys requires both the params and the point to be present */
-	if (!key->isPrivate && (found_params == NULL || found_point == NULL)) {
+	if (!key->isPrivate && (no_params || no_point)) {
 		EC_KEY_free(ec);
 		return NULL;
 	}
@@ -240,7 +260,16 @@ static EC_KEY *pkcs11_get_ec(PKCS11_KEY *key)
 	return ec;
 }
 
-static void pkcs11_set_ex_data_ec(EC_KEY* ec, PKCS11_KEY* key)
+static PKCS11_KEY *pkcs11_get_ex_data_ec(const EC_KEY *ec)
+{
+#if OPENSSL_VERSION_NUMBER >= 0x10100000L && !defined(LIBRESSL_VERSION_NUMBER)
+	return EC_KEY_get_ex_data(ec, ec_ex_index);
+#else
+	return ECDSA_get_ex_data((EC_KEY *)ec, ec_ex_index);
+#endif
+}
+
+static void pkcs11_set_ex_data_ec(EC_KEY *ec, PKCS11_KEY *key)
 {
 #if OPENSSL_VERSION_NUMBER >= 0x10100000L && !defined(LIBRESSL_VERSION_NUMBER)
 	EC_KEY_set_ex_data(ec, ec_ex_index, key);
@@ -249,15 +278,16 @@ static void pkcs11_set_ex_data_ec(EC_KEY* ec, PKCS11_KEY* key)
 #endif
 }
 
-static void pkcs11_update_ex_data_ec(PKCS11_KEY* key)
+static void pkcs11_update_ex_data_ec(PKCS11_KEY *key)
 {
-	EVP_PKEY* evp = key->evp_key;
+	EVP_PKEY *evp = key->evp_key;
+	EC_KEY *ec;
 	if (evp == NULL)
 		return;
 	if (EVP_PKEY_base_id(evp) != EVP_PKEY_EC)
 		return;
 
-	EC_KEY* ec = EVP_PKEY_get1_EC_KEY(evp);
+	ec = EVP_PKEY_get1_EC_KEY(evp);
 	pkcs11_set_ex_data_ec(ec, key);
 	EC_KEY_free(ec);
 }
@@ -325,7 +355,7 @@ static int pkcs11_ecdsa_sign(const unsigned char *msg, unsigned int msg_len,
 	CRYPTO_THREAD_write_lock(PRIVCTX(ctx)->rwlock);
 	rv = CRYPTOKI_call(ctx,
 		C_SignInit(spriv->session, &mechanism, kpriv->object));
-	if (rv == CKR_USER_NOT_LOGGED_IN)
+	if (!rv && kpriv->always_authenticate == CK_TRUE)
 		rv = pkcs11_authenticate(key);
 	if (!rv)
 		rv = CRYPTOKI_call(ctx,
@@ -333,7 +363,7 @@ static int pkcs11_ecdsa_sign(const unsigned char *msg, unsigned int msg_len,
 	CRYPTO_THREAD_unlock(PRIVCTX(ctx)->rwlock);
 
 	if (rv) {
-		PKCS11err(PKCS11_F_PKCS11_EC_KEY_SIGN, pkcs11_map_err(rv));
+		CKRerr(CKR_F_PKCS11_ECDSA_SIGN, rv);
 		return -1;
 	}
 	*siglen = ck_sigsize;
@@ -363,12 +393,8 @@ static ECDSA_SIG *pkcs11_ecdsa_sign_sig(const unsigned char *dgst, int dlen,
 	(void)kinv; /* Precomputed values are not used for PKCS#11 */
 	(void)rp; /* Precomputed values are not used for PKCS#11 */
 
-#if OPENSSL_VERSION_NUMBER >= 0x10100000L && !defined(LIBRESSL_VERSION_NUMBER)
-	key = (PKCS11_KEY *)EC_KEY_get_ex_data(ec, ec_ex_index);
-#else
-	key = (PKCS11_KEY *)ECDSA_get_ex_data(ec, ec_ex_index);
-#endif
-	if (key == NULL) {
+	key = pkcs11_get_ex_data_ec(ec);
+	if (check_key_fork(key) < 0) {
 		sign_sig_fn orig_sign_sig;
 #if OPENSSL_VERSION_NUMBER >= 0x10100000L && !defined(LIBRESSL_VERSION_NUMBER)
 		const EC_KEY_METHOD *meth = EC_KEY_OpenSSL();
@@ -380,7 +406,6 @@ static ECDSA_SIG *pkcs11_ecdsa_sign_sig(const unsigned char *dgst, int dlen,
 #endif
 		return orig_sign_sig(dgst, dlen, kinv, rp, ec);
 	}
-	/* TODO: Add an atfork check */
 
 	/* Truncate digest if its byte size is longer than needed */
 	order = BN_new();
@@ -509,18 +534,17 @@ static int pkcs11_ecdh_derive(unsigned char **out, size_t *outlen,
 			break;
 #endif
 		default:
-			PKCS11err(PKCS11_F_PKCS11_EC_KEY_COMPUTE_KEY, pkcs11_map_err(PKCS11_NOT_SUPPORTED));
+			P11err(P11_F_PKCS11_ECDH_DERIVE, P11_R_NOT_SUPPORTED);
 			return -1;
 	}
 
 	rv = CRYPTOKI_call(ctx, C_DeriveKey(spriv->session, &mechanism, kpriv->object, newkey_template, 5, &newkey));
-	CRYPTOKI_checkerr(PKCS11_F_PKCS11_EC_KEY_COMPUTE_KEY, rv);
+	CRYPTOKI_checkerr(CKR_F_PKCS11_ECDH_DERIVE, rv);
 
 	/* Return the value of the secret key and/or the object handle of the secret key */
 	if (out && outlen) { /* pkcs11_ec_ckey only asks for the value */
 		if (pkcs11_getattr_alloc(token, newkey, CKA_VALUE, out, outlen)) {
-			PKCS11err(PKCS11_F_PKCS11_EC_KEY_COMPUTE_KEY,
-				pkcs11_map_err(CKR_ATTRIBUTE_VALUE_INVALID));
+			CKRerr(CKR_F_PKCS11_ECDH_DERIVE, CKR_ATTRIBUTE_VALUE_INVALID);
 			CRYPTOKI_call(ctx, C_DestroyObject(spriv->session, newkey));
 			return -1;
 		}
@@ -554,10 +578,9 @@ static int pkcs11_ec_ckey(unsigned char **out, size_t *outlen,
 	size_t buflen;
 	int rv;
 
-	key = (PKCS11_KEY *)EC_KEY_get_ex_data(ecdh, ec_ex_index);
-	if (key == NULL) /* The private key is not handled by PKCS#11 */
+	key = pkcs11_get_ex_data_ec(ecdh);
+	if (check_key_fork(key) < 0)
 		return ossl_ecdh_compute_key(out, outlen, peer_point, ecdh);
-	/* TODO: Add an atfork check */
 
 	/* both peer and ecdh use same group parameters */
 	parms = pkcs11_ecdh_params_alloc(EC_KEY_get0_group(ecdh), peer_point);
@@ -596,14 +619,9 @@ static int pkcs11_ec_ckey(void *out, size_t outlen,
 	size_t buflen;
 	int rv;
 
-#if OPENSSL_VERSION_NUMBER >= 0x10100000L && !defined(LIBRESSL_VERSION_NUMBER)
-	key = (PKCS11_KEY *)EC_KEY_get_ex_data(ecdh, ec_ex_index);
-#else
-	key = (PKCS11_KEY *)ECDSA_get_ex_data((EC_KEY *)ecdh, ec_ex_index);
-#endif
-	if (key == NULL) /* The private key is not handled by PKCS#11 */
+	key = pkcs11_get_ex_data_ec(ecdh);
+	if (check_key_fork(key) < 0)
 		return ossl_ecdh_compute_key(out, outlen, peer_point, ecdh, KDF);
-	/* TODO: Add an atfork check */
 
 	/* both peer and ecdh use same group parameters */
 	parms = pkcs11_ecdh_params_alloc(EC_KEY_get0_group(ecdh), peer_point);
