@@ -1,6 +1,6 @@
 /* libp11, a simple layer on to of PKCS#11 API
  * Copyright (C) 2005 Olaf Kirch <okir@lst.de>
- * Copyright (C) 2016 Michał Trojnara <Michal.Trojnara@stunnel.org>
+ * Copyright (C) 2016-2018 Michał Trojnara <Michal.Trojnara@stunnel.org>
  *
  *  This library is free software; you can redistribute it and/or
  *  modify it under the terms of the GNU Lesser General Public
@@ -115,79 +115,85 @@ int pkcs11_reload_key(PKCS11_KEY *key)
 
 	rv = CRYPTOKI_call(ctx,
 		C_FindObjectsInit(spriv->session, key_search_attrs, 2));
-	CRYPTOKI_checkerr(PKCS11_F_PKCS11_ENUM_KEYS, rv);
+	CRYPTOKI_checkerr(CKR_F_PKCS11_RELOAD_KEY, rv);
 
 	rv = CRYPTOKI_call(ctx,
 		C_FindObjects(spriv->session, &kpriv->object, 1, &count));
-	CRYPTOKI_checkerr(PKCS11_F_PKCS11_ENUM_KEYS, rv);
+	CRYPTOKI_checkerr(CKR_F_PKCS11_RELOAD_KEY, rv);
 
 	CRYPTOKI_call(ctx, C_FindObjectsFinal(spriv->session));
 
 	return 0;
 }
 
-/*
- * Generate and store a private key on the token
- * FIXME: We should check first whether the token supports
- * on-board key generation, and if it does, use its own algorithm
+/**
+ * Generate a keyPair directly on token
  */
 int pkcs11_generate_key(PKCS11_TOKEN *token, int algorithm, unsigned int bits,
-		char *label, unsigned char* id, size_t id_len)
-{
-	PKCS11_KEY *key_obj;
-	EVP_PKEY *pk;
-	RSA *rsa;
-	BIO *err;
-#if OPENSSL_VERSION_NUMBER >= 0x10100000L && !defined(LIBRESSL_VERSION_NUMBER)
-	BIGNUM *exp;
-	BN_GENCB *gencb;
-#endif
-	int rc;
+		char *label, unsigned char* id, size_t id_len) {
 
-	if (algorithm != EVP_PKEY_RSA) {
-		PKCS11err(PKCS11_F_PKCS11_GENERATE_KEY, PKCS11_NOT_SUPPORTED);
+	PKCS11_SLOT *slot = TOKEN2SLOT(token);
+	PKCS11_CTX *ctx = TOKEN2CTX(token);
+	PKCS11_SLOT_private *spriv = PRIVSLOT(slot);
+
+	CK_ATTRIBUTE pubkey_attrs[32];
+	CK_ATTRIBUTE privkey_attrs[32];
+	unsigned int n_pub = 0, n_priv = 0;
+	CK_MECHANISM mechanism = {
+		CKM_RSA_PKCS_KEY_PAIR_GEN, NULL_PTR, 0
+	};
+	CK_BYTE public_exponent[] = { 1, 0, 1 };
+	CK_OBJECT_HANDLE pub_key_obj, priv_key_obj;
+	int rv;
+
+	(void)algorithm; /* squash the unused parameter warning */
+
+	/* make sure we have a session */
+	if (!spriv->haveSession && PKCS11_open_session(slot, 1))
 		return -1;
-	}
 
-	err = BIO_new_fp(stderr, BIO_NOCLOSE);
+	/* pubkey attributes */
+	pkcs11_addattr(pubkey_attrs + n_pub++, CKA_ID, id, id_len);
+	if (label)
+		pkcs11_addattr_s(pubkey_attrs + n_pub++, CKA_LABEL, label);
+	pkcs11_addattr_bool(pubkey_attrs + n_pub++, CKA_TOKEN, TRUE);
 
-#if OPENSSL_VERSION_NUMBER >= 0x10100000L && !defined(LIBRESSL_VERSION_NUMBER)
-	exp = BN_new();
-	rsa = RSA_new();
-	gencb = BN_GENCB_new();
-	if (gencb)
-		BN_GENCB_set(gencb, NULL, err);
-	if (rsa == NULL || exp == NULL || gencb == NULL ||
-			!BN_set_word(exp, RSA_F4) ||
-			!RSA_generate_key_ex(rsa, bits, exp, gencb)) {
-		RSA_free(rsa);
-		rsa = NULL;
-	}
-	BN_GENCB_free(gencb);
-	BN_free(exp);
-#else
-	rsa = RSA_generate_key(bits, RSA_F4, NULL, err);
-#endif
-	BIO_free(err);
-	if (rsa == NULL) {
-		PKCS11err(PKCS11_F_PKCS11_GENERATE_KEY, PKCS11_KEYGEN_FAILED);
-		return -1;
-	}
+	pkcs11_addattr_bool(pubkey_attrs + n_pub++, CKA_ENCRYPT, TRUE);
+	pkcs11_addattr_bool(pubkey_attrs + n_pub++, CKA_VERIFY, TRUE);
+	pkcs11_addattr_bool(pubkey_attrs + n_pub++, CKA_WRAP, TRUE);
+	pkcs11_addattr_int(pubkey_attrs + n_pub++, CKA_MODULUS_BITS, bits);
+	pkcs11_addattr(pubkey_attrs + n_pub++, CKA_PUBLIC_EXPONENT, public_exponent, 3);
 
-	pk = EVP_PKEY_new();
-	EVP_PKEY_assign_RSA(pk, rsa);
-	rc = pkcs11_store_key(token, pk, CKO_PRIVATE_KEY,
-		label, id, id_len, &key_obj);
+	/* privkey attributes */
+	pkcs11_addattr(privkey_attrs + n_priv++, CKA_ID, id, id_len);
+	if (label)
+		pkcs11_addattr_s(privkey_attrs + n_priv++, CKA_LABEL, label);
+	pkcs11_addattr_bool(privkey_attrs + n_priv++, CKA_TOKEN, TRUE);
+	pkcs11_addattr_bool(privkey_attrs + n_priv++, CKA_PRIVATE, TRUE);
+	pkcs11_addattr_bool(privkey_attrs + n_priv++, CKA_SENSITIVE, TRUE);
+	pkcs11_addattr_bool(privkey_attrs + n_priv++, CKA_DECRYPT, TRUE);
+	pkcs11_addattr_bool(privkey_attrs + n_priv++, CKA_SIGN, TRUE);
+	pkcs11_addattr_bool(privkey_attrs + n_priv++, CKA_UNWRAP, TRUE);
 
-	if (rc == 0) {
-		PKCS11_KEY_private *kpriv;
+	/* call the pkcs11 module to create the key pair */
+	rv = CRYPTOKI_call(ctx, C_GenerateKeyPair(
+		spriv->session,
+		&mechanism,
+		pubkey_attrs,
+		n_pub,
+		privkey_attrs,
+		n_priv,
+		&pub_key_obj,
+		&priv_key_obj
+	));
 
-		kpriv = PRIVKEY(key_obj);
-		rc = pkcs11_store_key(token, pk, CKO_PUBLIC_KEY,
-			label, kpriv->id, kpriv->id_len, NULL);
-	}
-	EVP_PKEY_free(pk);
-	return rc;
+	/* zap all memory allocated when building the template */
+	pkcs11_zap_attrs(privkey_attrs, n_priv);
+	pkcs11_zap_attrs(pubkey_attrs, n_pub);
+
+	CRYPTOKI_checkerr(CKR_F_PKCS11_GENERATE_KEY, rv);
+
+	return 0;
 }
 
 /*
@@ -223,7 +229,7 @@ static int pkcs11_store_key(PKCS11_TOKEN *token, EVP_PKEY *pk,
 	CK_ATTRIBUTE attrs[32];
 	unsigned int n = 0;
 	int rv;
-	const BIGNUM *rsa_n, *rsa_e, *rsa_d, *rsa_p, *rsa_q;
+	const BIGNUM *rsa_n, *rsa_e, *rsa_d, *rsa_p, *rsa_q, *rsa_dmp1, *rsa_dmq1, *rsa_iqmp;
 
 	/* First, make sure we have a session */
 	if (!spriv->haveSession && PKCS11_open_session(slot, 1))
@@ -250,20 +256,23 @@ static int pkcs11_store_key(PKCS11_TOKEN *token, EVP_PKEY *pk,
 #if OPENSSL_VERSION_NUMBER >= 0x10100003L && !defined(LIBRESSL_VERSION_NUMBER)
 	if (EVP_PKEY_base_id(pk) == EVP_PKEY_RSA) {
 		RSA *rsa = EVP_PKEY_get1_RSA(pk);
+		pkcs11_addattr_int(attrs + n++, CKA_KEY_TYPE, CKK_RSA);
+		RSA_get0_key(rsa, &rsa_n, &rsa_e, &rsa_d);
+		RSA_get0_factors(rsa, &rsa_p, &rsa_q);
+		RSA_get0_crt_params(rsa, &rsa_dmp1, &rsa_dmq1, &rsa_iqmp);
+		RSA_free(rsa);
 #else
 	if (pk->type == EVP_PKEY_RSA) {
 		RSA *rsa = pk->pkey.rsa;
-#endif
 		pkcs11_addattr_int(attrs + n++, CKA_KEY_TYPE, CKK_RSA);
-#if OPENSSL_VERSION_NUMBER >= 0x10100005L && !defined(LIBRESSL_VERSION_NUMBER)
-		RSA_get0_key(rsa, &rsa_n, &rsa_e, &rsa_d);
-		RSA_get0_factors(rsa, &rsa_p, &rsa_q);
-#else
 		rsa_n=rsa->n;
 		rsa_e=rsa->e;
 		rsa_d=rsa->d;
 		rsa_p=rsa->p;
 		rsa_q=rsa->q;
+		rsa_dmp1=rsa->dmp1;
+		rsa_dmq1=rsa->dmq1;
+		rsa_iqmp=rsa->iqmp;
 #endif
 		pkcs11_addattr_bn(attrs + n++, CKA_MODULUS, rsa_n);
 		pkcs11_addattr_bn(attrs + n++, CKA_PUBLIC_EXPONENT, rsa_e);
@@ -271,13 +280,16 @@ static int pkcs11_store_key(PKCS11_TOKEN *token, EVP_PKEY *pk,
 			pkcs11_addattr_bn(attrs + n++, CKA_PRIVATE_EXPONENT, rsa_d);
 			pkcs11_addattr_bn(attrs + n++, CKA_PRIME_1, rsa_p);
 			pkcs11_addattr_bn(attrs + n++, CKA_PRIME_2, rsa_q);
+			if (rsa_dmp1)
+				pkcs11_addattr_bn(attrs + n++, CKA_EXPONENT_1, rsa_dmp1);
+			if (rsa_dmq1)
+				pkcs11_addattr_bn(attrs + n++, CKA_EXPONENT_2, rsa_dmq1);
+			if (rsa_iqmp)
+				pkcs11_addattr_bn(attrs + n++, CKA_COEFFICIENT, rsa_iqmp);
 		}
 	} else {
 		pkcs11_zap_attrs(attrs, n);
-		PKCS11err(type == CKO_PRIVATE_KEY ?
-				PKCS11_F_PKCS11_STORE_PRIVATE_KEY :
-				PKCS11_F_PKCS11_STORE_PUBLIC_KEY,
-			PKCS11_NOT_SUPPORTED);
+		P11err(P11_F_PKCS11_STORE_KEY, P11_R_NOT_SUPPORTED);
 		return -1;
 	}
 
@@ -287,7 +299,7 @@ static int pkcs11_store_key(PKCS11_TOKEN *token, EVP_PKEY *pk,
 	/* Zap all memory allocated when building the template */
 	pkcs11_zap_attrs(attrs, n);
 
-	CRYPTOKI_checkerr(PKCS11_F_PKCS11_STORE_PRIVATE_KEY, rv);
+	CRYPTOKI_checkerr(CKR_F_PKCS11_STORE_KEY, rv);
 
 	/* Gobble the key object */
 	return pkcs11_init_key(ctx, token, spriv->session, object, type, ret_key);
@@ -335,6 +347,7 @@ EVP_PKEY *pkcs11_get_key(PKCS11_KEY *key, int isPrivate)
 
 /*
  * Authenticate a private the key operation if needed
+ * This function *only* handles CKU_CONTEXT_SPECIFIC logins.
  */
 int pkcs11_authenticate(PKCS11_KEY *key)
 {
@@ -358,25 +371,25 @@ int pkcs11_authenticate(PKCS11_KEY *key)
 	/* Call UI to ask for a PIN */
 	ui = UI_new_method(cpriv->ui_method);
 	if (ui == NULL)
-		return PKCS11_UI_FAILED;
+		return P11_R_UI_FAILED;
 	if (cpriv->ui_user_data != NULL)
 		UI_add_user_data(ui, cpriv->ui_user_data);
 	memset(pin, 0, MAX_PIN_LENGTH+1);
 	prompt = UI_construct_prompt(ui, "PKCS#11 key PIN", key->label);
 	if (!prompt) {
-		return PKCS11_UI_FAILED;
+		return P11_R_UI_FAILED;
 	}
 	if (!UI_dup_input_string(ui, prompt,
 			UI_INPUT_FLAG_DEFAULT_PWD, pin, 4, MAX_PIN_LENGTH)) {
 		UI_free(ui);
 		OPENSSL_free(prompt);
-		return PKCS11_UI_FAILED;
+		return P11_R_UI_FAILED;
 	}
 	OPENSSL_free(prompt);
 
 	if (UI_process(ui)) {
 		UI_free(ui);
-		return PKCS11_UI_FAILED;
+		return P11_R_UI_FAILED;
 	}
 	UI_free(ui);
 
@@ -433,6 +446,51 @@ int pkcs11_enumerate_keys(PKCS11_TOKEN *token, unsigned int type,
 	return 0;
 }
 
+/**
+ * Remove a key from the associated token
+ */ 
+int pkcs11_remove_key(PKCS11_KEY *key) {
+	PKCS11_SLOT *slot = KEY2SLOT(key);
+	PKCS11_CTX *ctx = KEY2CTX(key);
+	PKCS11_SLOT_private *spriv = PRIVSLOT(slot);
+	CK_OBJECT_HANDLE obj;
+	CK_ULONG count;
+	CK_ATTRIBUTE search_parameters[32];
+	unsigned int n = 0;
+
+	/* First, make sure we have a session */
+	if (!spriv->haveSession && PKCS11_open_session(slot, 1))
+		return -1;
+	if (key->isPrivate)
+		pkcs11_addattr_int(search_parameters + n++, CKA_CLASS, CKO_PRIVATE_KEY);
+	else
+		pkcs11_addattr_int(search_parameters + n++, CKA_CLASS, CKO_PUBLIC_KEY);
+	if (key->id && key->id_len)
+		pkcs11_addattr(search_parameters + n++, CKA_ID, key->id, key->id_len);
+	if (key->label)
+	 	pkcs11_addattr_s(search_parameters + n++, CKA_LABEL, key->label);
+
+	int rv = CRYPTOKI_call(ctx,
+		C_FindObjectsInit(spriv->session, search_parameters, n));
+	CRYPTOKI_checkerr(CKR_F_PKCS11_REMOVE_KEY, rv);
+
+	rv = CRYPTOKI_call(ctx, C_FindObjects(spriv->session, &obj, 1, &count));
+	CRYPTOKI_checkerr(CKR_F_PKCS11_REMOVE_KEY, rv);
+
+	CRYPTOKI_call(ctx, C_FindObjectsFinal(spriv->session));
+	if (count!=1) {
+		pkcs11_zap_attrs(search_parameters, n);
+		return -1;
+	}
+	rv = CRYPTOKI_call(ctx, C_DestroyObject(spriv->session, obj));
+	if (rv != CKR_OK) {
+		pkcs11_zap_attrs(search_parameters, n);
+		return -1;
+	}
+	pkcs11_zap_attrs(search_parameters, n);
+	return 0;
+}
+
 /*
  * Find all keys of a given type (public or private)
  */
@@ -451,7 +509,7 @@ static int pkcs11_find_keys(PKCS11_TOKEN *token, unsigned int type)
 	key_search_class = type;
 	rv = CRYPTOKI_call(ctx,
 		C_FindObjectsInit(spriv->session, key_search_attrs, 1));
-	CRYPTOKI_checkerr(PKCS11_F_PKCS11_ENUM_KEYS, rv);
+	CRYPTOKI_checkerr(CKR_F_PKCS11_FIND_KEYS, rv);
 
 	do {
 		res = pkcs11_next_key(ctx, token, spriv->session, type);
@@ -471,7 +529,7 @@ static int pkcs11_next_key(PKCS11_CTX *ctx, PKCS11_TOKEN *token,
 
 	/* Get the next matching object */
 	rv = CRYPTOKI_call(ctx, C_FindObjects(session, &obj, 1, &count));
-	CRYPTOKI_checkerr(PKCS11_F_PKCS11_ENUM_KEYS, rv);
+	CRYPTOKI_checkerr(CKR_F_PKCS11_NEXT_KEY, rv);
 
 	if (count == 0)
 		return 1;
@@ -549,7 +607,7 @@ static int pkcs11_init_key(PKCS11_CTX *ctx, PKCS11_TOKEN *token,
 	if (pkcs11_getattr_var(token, obj, CKA_ID, kpriv->id, &kpriv->id_len))
 		kpriv->id_len = 0;
 	kpriv->ops = ops;
-	kpriv->forkid = _P11_get_forkid();
+	kpriv->forkid = get_forkid();
 
 	if (ret)
 		*ret = key;
