@@ -39,8 +39,10 @@ PKCS11_CTX *pkcs11_CTX_new(void)
 		goto fail;
 	memset(ctx, 0, sizeof(PKCS11_CTX));
 	ctx->_private = cpriv;
-	cpriv->forkid = _P11_get_forkid();
+	cpriv->forkid = get_forkid();
 	cpriv->rwlock = CRYPTO_THREAD_lock_new();
+	cpriv->sign_initialized = 0;
+	cpriv->decrypt_initialized = 0;
 
 	return ctx;
 fail:
@@ -52,7 +54,7 @@ fail:
 /*
  * Set private init args for module
  */
-void pkcs11_CTX_init_args(PKCS11_CTX * ctx, const char *init_args)
+void pkcs11_CTX_init_args(PKCS11_CTX *ctx, const char *init_args)
 {
 	PKCS11_CTX_private *cpriv = PRIVCTX(ctx);
 	/* Free previously duplicated string */
@@ -65,7 +67,7 @@ void pkcs11_CTX_init_args(PKCS11_CTX * ctx, const char *init_args)
 /*
  * Load the shared library, and initialize it.
  */
-int pkcs11_CTX_load(PKCS11_CTX * ctx, const char *name)
+int pkcs11_CTX_load(PKCS11_CTX *ctx, const char *name)
 {
 	PKCS11_CTX_private *cpriv = PRIVCTX(ctx);
 	CK_C_INITIALIZE_ARGS args;
@@ -74,7 +76,7 @@ int pkcs11_CTX_load(PKCS11_CTX * ctx, const char *name)
 
 	cpriv->handle = C_LoadModule(name, &cpriv->method);
 	if (cpriv->handle == NULL) {
-		PKCS11err(PKCS11_F_PKCS11_CTX_LOAD, PKCS11_LOAD_MODULE_ERROR);
+		P11err(P11_F_PKCS11_CTX_LOAD, P11_R_LOAD_MODULE_ERROR);
 		return -1;
 	}
 
@@ -85,13 +87,21 @@ int pkcs11_CTX_load(PKCS11_CTX * ctx, const char *name)
 	args.pReserved = cpriv->init_args;
 	rv = cpriv->method->C_Initialize(&args);
 	if (rv && rv != CKR_CRYPTOKI_ALREADY_INITIALIZED) {
-		PKCS11err(PKCS11_F_PKCS11_CTX_LOAD, rv);
+		C_UnloadModule(cpriv->handle);
+		cpriv->handle = NULL;
+		CKRerr(P11_F_PKCS11_CTX_LOAD, rv);
 		return -1;
 	}
 
 	/* Get info on the library */
 	rv = cpriv->method->C_GetInfo(&ck_info);
-	CRYPTOKI_checkerr(PKCS11_F_PKCS11_CTX_LOAD, rv);
+	if (rv) {
+		cpriv->method->C_Finalize(NULL);
+		C_UnloadModule(cpriv->handle);
+		cpriv->handle = NULL;
+		CKRerr(P11_F_PKCS11_CTX_LOAD, rv);
+		return -1;
+	}
 
 	ctx->manufacturer = PKCS11_DUP(ck_info.manufacturerID);
 	ctx->description = PKCS11_DUP(ck_info.libraryDescription);
@@ -102,7 +112,7 @@ int pkcs11_CTX_load(PKCS11_CTX * ctx, const char *name)
 /*
  * Reinitialize (e.g., after a fork).
  */
-int pkcs11_CTX_reload(PKCS11_CTX * ctx)
+int pkcs11_CTX_reload(PKCS11_CTX *ctx)
 {
 	PKCS11_CTX_private *cpriv = PRIVCTX(ctx);
 	CK_C_INITIALIZE_ARGS _args;
@@ -120,34 +130,33 @@ int pkcs11_CTX_reload(PKCS11_CTX * ctx)
 	}
 	rv = cpriv->method->C_Initialize(args);
 	if (rv && rv != CKR_CRYPTOKI_ALREADY_INITIALIZED) {
-		PKCS11err(PKCS11_F_PKCS11_CTX_LOAD, rv);
+		CKRerr(P11_F_PKCS11_CTX_RELOAD, rv);
 		return -1;
 	}
 
-	/* Reinitialize the PKCS11 internal slot table */
-	return pkcs11_enumerate_slots(ctx, NULL, NULL);
+	return 0;
 }
 
 /*
  * Unload the shared library
  */
-void pkcs11_CTX_unload(PKCS11_CTX * ctx)
+void pkcs11_CTX_unload(PKCS11_CTX *ctx)
 {
-	PKCS11_CTX_private *cpriv;
-	cpriv = PRIVCTX(ctx);
+	PKCS11_CTX_private *cpriv = PRIVCTX(ctx);
 
 	/* Tell the PKCS11 library to shut down */
-	if (cpriv->forkid == _P11_get_forkid())
+	if (cpriv->forkid == get_forkid())
 		cpriv->method->C_Finalize(NULL);
 
 	/* Unload the module */
 	C_UnloadModule(cpriv->handle);
+	cpriv->handle = NULL;
 }
 
 /*
  * Free a context
  */
-void pkcs11_CTX_free(PKCS11_CTX * ctx)
+void pkcs11_CTX_free(PKCS11_CTX *ctx)
 {
 	PKCS11_CTX_private *cpriv = PRIVCTX(ctx);
 
@@ -159,6 +168,9 @@ void pkcs11_CTX_free(PKCS11_CTX * ctx)
 	if (cpriv->init_args) {
 		OPENSSL_free(cpriv->init_args);
 	}
+	if(cpriv->handle) {
+		OPENSSL_free(cpriv->handle);
+	}
 	CRYPTO_THREAD_lock_free(cpriv->rwlock);
 	OPENSSL_free(ctx->manufacturer);
 	OPENSSL_free(ctx->description);
