@@ -1,6 +1,6 @@
 /* libp11, a simple layer on to of PKCS#11 API
  * Copyright (C) 2005 Olaf Kirch <okir@lst.de>
- * Copyright (C) 2016 Michał Trojnara <Michal.Trojnara@stunnel.org>
+ * Copyright (C) 2016-2017 Michał Trojnara <Michal.Trojnara@stunnel.org>
  *
  *  This library is free software; you can redistribute it and/or
  *  modify it under the terms of the GNU Lesser General Public
@@ -29,10 +29,6 @@
 
 static int rsa_ex_index = 0;
 
-#if OPENSSL_VERSION_NUMBER < 0x10100003L || defined(LIBRESSL_VERSION_NUMBER)
-#define EVP_PKEY_get0_RSA(key) ((key)->pkey.rsa)
-#endif
-
 static RSA *pkcs11_rsa(PKCS11_KEY *key)
 {
 	EVP_PKEY *evp_key = pkcs11_get_key(key, key->isPrivate);
@@ -61,8 +57,8 @@ static int pkcs11_mechanism(CK_MECHANISM *mechanism, const int padding)
 	memset(mechanism, 0, sizeof(CK_MECHANISM));
 	switch (padding) {
 	case RSA_PKCS1_PADDING:
-		mechanism->mechanism = CKM_RSA_PKCS;
-		break;
+		 mechanism->mechanism = CKM_RSA_PKCS;
+		 break;
 	case RSA_NO_PADDING:
 		mechanism->mechanism = CKM_RSA_X_509;
 		break;
@@ -70,7 +66,7 @@ static int pkcs11_mechanism(CK_MECHANISM *mechanism, const int padding)
 		mechanism->mechanism = CKM_RSA_X9_31;
 		break;
 	default:
-		fprintf(stderr, "PKCS#11: Unsupported padding type\n");
+		P11err(P11_F_PKCS11_MECHANISM, P11_R_UNSUPPORTED_PADDING_TYPE);
 		return -1;
 	}
 	return 0;
@@ -99,7 +95,7 @@ int pkcs11_private_encrypt(int flen,
 	/* Try signing first, as applications are more likely to use it */
 	rv = CRYPTOKI_call(ctx,
 		C_SignInit(spriv->session, &mechanism, kpriv->object));
-	if (rv == CKR_USER_NOT_LOGGED_IN)
+	if (!rv && kpriv->always_authenticate == CK_TRUE)
 		rv = pkcs11_authenticate(key);
 	if (!rv)
 		rv = CRYPTOKI_call(ctx,
@@ -108,7 +104,7 @@ int pkcs11_private_encrypt(int flen,
 		/* OpenSSL may use it for encryption rather than signing */
 		rv = CRYPTOKI_call(ctx,
 			C_EncryptInit(spriv->session, &mechanism, kpriv->object));
-		if (rv == CKR_USER_NOT_LOGGED_IN)
+		if (!rv && kpriv->always_authenticate == CK_TRUE)
 			rv = pkcs11_authenticate(key);
 		if (!rv)
 			rv = CRYPTOKI_call(ctx,
@@ -117,7 +113,7 @@ int pkcs11_private_encrypt(int flen,
 	CRYPTO_THREAD_unlock(PRIVCTX(ctx)->rwlock);
 
 	if (rv) {
-		PKCS11err(PKCS11_F_PKCS11_RSA_ENCRYPT, pkcs11_map_err(rv));
+		CKRerr(CKR_F_PKCS11_PRIVATE_ENCRYPT, rv);
 		return -1;
 	}
 
@@ -142,7 +138,7 @@ int pkcs11_private_decrypt(int flen, const unsigned char *from, unsigned char *t
 	CRYPTO_THREAD_write_lock(PRIVCTX(ctx)->rwlock);
 	rv = CRYPTOKI_call(ctx,
 		C_DecryptInit(spriv->session, &mechanism, kpriv->object));
-	if (rv == CKR_USER_NOT_LOGGED_IN)
+	if (!rv && kpriv->always_authenticate == CK_TRUE)
 		rv = pkcs11_authenticate(key);
 	if (!rv)
 		rv = CRYPTOKI_call(ctx,
@@ -151,7 +147,7 @@ int pkcs11_private_decrypt(int flen, const unsigned char *from, unsigned char *t
 	CRYPTO_THREAD_unlock(PRIVCTX(ctx)->rwlock);
 
 	if (rv) {
-		PKCS11err(PKCS11_F_PKCS11_RSA_DECRYPT, pkcs11_map_err(rv));
+		CKRerr(CKR_F_PKCS11_PRIVATE_DECRYPT, rv);
 		return -1;
 	}
 
@@ -170,7 +166,7 @@ int pkcs11_verify(int type, const unsigned char *m, unsigned int m_len,
 	(void)key;
 
 	/* PKCS11 calls go here */
-	PKCS11err(PKCS11_F_PKCS11_RSA_VERIFY, PKCS11_NOT_SUPPORTED);
+	P11err(P11_F_PKCS11_VERIFY, P11_R_NOT_SUPPORTED);
 	return -1;
 }
 
@@ -180,38 +176,33 @@ int pkcs11_verify(int type, const unsigned char *m, unsigned int m_len,
 static RSA *pkcs11_get_rsa(PKCS11_KEY *key)
 {
 	RSA *rsa;
-	PKCS11_KEY *keys = NULL;
-	unsigned int i, count = 0;
-	BIGNUM *rsa_n=NULL, *rsa_e=NULL;
+	PKCS11_KEY *keys;
+	unsigned int i, count;
+	BIGNUM *rsa_n = NULL, *rsa_e = NULL;
 
-	rsa = RSA_new();
-	if (rsa == NULL)
+	/* Retrieve the modulus */
+	if (key_getattr_bn(key, CKA_MODULUS, &rsa_n))
 		return NULL;
 
-	/* Retrieve the modulus and the public exponent */
-	if (key_getattr_bn(key, CKA_MODULUS, &rsa_n) ||
-			key_getattr_bn(key, CKA_PUBLIC_EXPONENT, &rsa_e))
-		goto failure;
-	if (!BN_is_zero(rsa_e)) /* The public exponent was retrieved */
-		goto success;
-	BN_clear_free(rsa_e);
-	rsa_e = NULL;
+	/* Retrieve the public exponent */
+	if (!key_getattr_bn(key, CKA_PUBLIC_EXPONENT, &rsa_e)) {
+		if (!BN_is_zero(rsa_e)) /* A valid public exponent */
+			goto success;
+		BN_clear_free(rsa_e);
+		rsa_e = NULL;
+	}
 
 	/* The public exponent was not found in the private key:
 	 * retrieve it from the corresponding public key */
 	if (!PKCS11_enumerate_public_keys(KEY2TOKEN(key), &keys, &count)) {
 		for(i = 0; i < count; i++) {
-			BIGNUM *pubmod;
-
-			if (key_getattr_bn(&keys[i], CKA_MODULUS, &pubmod))
-				continue; /* Failed to retrieve the modulus */
-			if (BN_cmp(rsa_n, pubmod) == 0) { /* The key was found */
-				BN_clear_free(pubmod);
-				if (key_getattr_bn(&keys[i], CKA_PUBLIC_EXPONENT, &rsa_e))
-					continue; /* Failed to retrieve the public exponent */
-				goto success;
-			} else {
+			BIGNUM *pubmod = NULL;
+			if (!key_getattr_bn(&keys[i], CKA_MODULUS, &pubmod)) {
+				int found = BN_cmp(rsa_n, pubmod) == 0;
 				BN_clear_free(pubmod);
+				if (found && !key_getattr_bn(&keys[i],
+						CKA_PUBLIC_EXPONENT, &rsa_e))
+					goto success;
 			}
 		}
 	}
@@ -222,33 +213,45 @@ static RSA *pkcs11_get_rsa(PKCS11_KEY *key)
 		goto success;
 
 failure:
-	RSA_free(rsa);
+	if (rsa_n)
+		BN_clear_free(rsa_n);
+	if (rsa_e)
+		BN_clear_free(rsa_e);
 	return NULL;
 
 success:
+	rsa = RSA_new();
+	if (rsa == NULL)
+		goto failure;
 #if OPENSSL_VERSION_NUMBER >= 0x10100005L && !defined(LIBRESSL_VERSION_NUMBER)
-		RSA_set0_key(rsa, rsa_n, rsa_e, NULL);
+	RSA_set0_key(rsa, rsa_n, rsa_e, NULL);
 #else
-		rsa->n=rsa_n;
-		rsa->e=rsa_e;
+	rsa->n = rsa_n;
+	rsa->e = rsa_e;
 #endif
 	return rsa;
 }
 
-static void pkcs11_set_ex_data_rsa(RSA* rsa, PKCS11_KEY* key)
+PKCS11_KEY *pkcs11_get_ex_data_rsa(const RSA *rsa)
+{
+	return RSA_get_ex_data(rsa, rsa_ex_index);
+}
+
+static void pkcs11_set_ex_data_rsa(RSA *rsa, PKCS11_KEY *key)
 {
 	RSA_set_ex_data(rsa, rsa_ex_index, key);
 }
 
-static void pkcs11_update_ex_data_rsa(PKCS11_KEY* key)
+static void pkcs11_update_ex_data_rsa(PKCS11_KEY *key)
 {
-	EVP_PKEY* evp = key->evp_key;
+	EVP_PKEY *evp = key->evp_key;
+	RSA *rsa;
 	if (evp == NULL)
 		return;
 	if (EVP_PKEY_base_id(evp) != EVP_PKEY_RSA)
 		return;
 
-	RSA* rsa = EVP_PKEY_get1_RSA(evp);
+	rsa = EVP_PKEY_get1_RSA(evp);
 	pkcs11_set_ex_data_rsa(rsa, key);
 	RSA_free(rsa);
 }
@@ -348,10 +351,10 @@ int (*RSA_meth_get_priv_dec(const RSA_METHOD *meth))
 static int pkcs11_rsa_priv_dec_method(int flen, const unsigned char *from,
 		unsigned char *to, RSA *rsa, int padding)
 {
-	PKCS11_KEY *key = RSA_get_ex_data(rsa, rsa_ex_index);
+	PKCS11_KEY *key = pkcs11_get_ex_data_rsa(rsa);
 	int (*priv_dec) (int flen, const unsigned char *from,
 		unsigned char *to, RSA *rsa, int padding);
-	if (key == NULL) {
+	if (check_key_fork(key) < 0) {
 		priv_dec = RSA_meth_get_priv_dec(RSA_get_default_method());
 		return priv_dec(flen, from, to, rsa, padding);
 	}
@@ -361,10 +364,10 @@ static int pkcs11_rsa_priv_dec_method(int flen, const unsigned char *from,
 static int pkcs11_rsa_priv_enc_method(int flen, const unsigned char *from,
 		unsigned char *to, RSA *rsa, int padding)
 {
-	PKCS11_KEY *key = RSA_get_ex_data(rsa, rsa_ex_index);
+	PKCS11_KEY *key = pkcs11_get_ex_data_rsa(rsa);
 	int (*priv_enc) (int flen, const unsigned char *from,
 		unsigned char *to, RSA *rsa, int padding);
-	if (key == NULL) {
+	if (check_key_fork(key) < 0) {
 		priv_enc = RSA_meth_get_priv_enc(RSA_get_default_method());
 		return priv_enc(flen, from, to, rsa, padding);
 	}
